package database

import (
	"chromagies/src/config"
	"chromagies/src/profiler"
	"database/sql"
	"fmt"
	"log"
	"strings"
	"time"

	_ "github.com/go-sql-driver/mysql"
)

type DatabaseError struct {
	message string
}

func createDatabaseError(message string) DatabaseError {
	log.Printf("Database error : %s\n", message)
	return DatabaseError{
		message: message,
	}
}

func (e *DatabaseError) Error() string {
	return e.message
}

func (e *DatabaseError) HasError() bool {
	return e.message != ""
}

type DatabaseResult struct {
	Error        DatabaseError
	SqlRows      *sql.Rows
	SqlResult    sql.Result
	ParsedResult map[string][]sql.NullString
}

const (
	profiler_DATABASE_CONNECT string = "Database Connect"
	profiler_DATABASE_QUERY   string = "Database Query"
)

const (
	database_ROUTINE_TYPE_PROCEDURE uint8 = 1
	database_ROUTINE_TYPE_FUNCTION  uint8 = 2
)

type databaseEnumIndexType string

const (
	database_ENUM_INDEX_PRIMARY databaseEnumIndexType = "PRIMARY"
	database_ENUM_INDEX_UNIQUE  databaseEnumIndexType = "UNIQUE"
)

type columsDefinition struct {
	Name          string
	Type          string
	Default       any
	IsNullable    bool
	AutoIncrement bool
	Charset       any
	Collation     any
	Table         string
	Primary       bool
}

type indexDefinition struct {
	Type    databaseEnumIndexType
	Name    string
	Table   string
	Columns []string
}

type foreignKeyDefinition struct {
	Name             string
	Table            string
	ColumnName       string
	PointingToTable  string
	PointingToColumn string
	UpdateRule       string
	DeleteRule       string
}

type tableDefinition struct {
	Name    string
	Columns []columsDefinition
}

type storedRoutineParameterDefinition struct {
	Name       string
	Type       string
	IsNullable bool
	Charset    any
	Collation  any
}

type storedFunctionDefinition struct {
	Name       string
	Parameters []storedRoutineParameterDefinition
	Return     storedRoutineParameterDefinition
	Command    string
}

type storedProcedureDefinition struct {
	Name       string
	Parameters []storedRoutineParameterDefinition
	Command    string
}

type databaseDefinition struct {
	Tables           []tableDefinition
	Indexes          []indexDefinition
	ForeignKeys      []foreignKeyDefinition
	StoredProcedures []storedProcedureDefinition
	StoredFunctions  []storedFunctionDefinition
}

var databaseStructure databaseDefinition = databaseDefinition{
	Tables: []tableDefinition{
		{
			Name: "user",
			Columns: []columsDefinition{
				{
					Name:          "id",
					Type:          "int(11)",
					IsNullable:    false,
					AutoIncrement: true,
					Primary:       true,
				},
				{
					Name:       "name",
					Type:       "varchar(256)",
					IsNullable: false,
					Charset:    "utf8mb4",
					Collation:  "utf8mb4_unicode_ci",
				},
				{
					Name:       "password",
					Type:       "varchar(256)",
					IsNullable: false,
					Charset:    "utf8mb4",
					Collation:  "utf8mb4_bin",
				},
				{
					Name:       "password_salt",
					Type:       "varchar(256)",
					IsNullable: false,
					Charset:    "utf8mb4",
					Collation:  "utf8mb4_bin",
				},
				{
					Name:       "permission_level",
					Type:       "enum('ADMIN', 'MODERATOR', 'USER')",
					Default:    "'USER'",
					IsNullable: false,
					Charset:    nil,
				},
				{
					Name:       "email",
					Type:       "varchar(256)",
					IsNullable: true,
					Default:    "NULL",
					Charset:    "utf8mb4",
					Collation:  "utf8mb4_bin",
				},
			},
		},
		{
			Name: "folder",
			Columns: []columsDefinition{
				{
					Name:          "id",
					Type:          "int(11)",
					IsNullable:    false,
					AutoIncrement: true,
					Primary:       true,
				},
				{
					Name:       "name",
					Type:       "varchar(64)",
					IsNullable: false,
					Charset:    "utf8mb4",
					Collation:  "utf8mb4_unicode_ci",
				},
			},
		},
		{
			Name: "tag",
			Columns: []columsDefinition{
				{
					Name:          "id",
					Type:          "int(11)",
					IsNullable:    false,
					AutoIncrement: true,
					Primary:       true,
				},
				{
					Name:       "name",
					Type:       "varchar(64)",
					IsNullable: false,
					Charset:    "utf8mb4",
					Collation:  "utf8mb4_unicode_ci",
				},
			},
		},
		{
			Name: "website",
			Columns: []columsDefinition{
				{
					Name:          "id",
					Type:          "int(11)",
					IsNullable:    false,
					AutoIncrement: true,
					Primary:       true,
				},
				{
					Name:       "name",
					Type:       "varchar(256)",
					IsNullable: false,
					Charset:    "utf8mb4",
					Collation:  "utf8mb4_unicode_ci",
				},
				{
					Name:       "content",
					Type:       "longtext",
					IsNullable: true,
					Default:    "NULL",
					Charset:    "utf8mb4",
					Collation:  "utf8mb4_bin",
				},
				{
					Name:       "description",
					Type:       "longtext",
					IsNullable: true,
					Default:    "NULL",
					Charset:    "utf8mb4",
					Collation:  "utf8mb4_bin",
				},
				{
					Name:       "folder_id",
					Type:       "int(11)",
					IsNullable: false,
				},
				{
					Name:       "date",
					Type:       "datetime",
					IsNullable: false,
					Default:    "current_timestamp()",
				},
				{
					Name:       "title",
					Type:       "varchar(256)",
					IsNullable: false,
					Charset:    "utf8mb4",
					Collation:  "utf8mb4_unicode_ci",
				},
			},
		},
		{
			Name: "tags_websites",
			Columns: []columsDefinition{
				{
					Name:       "tag_id",
					Type:       "int(11)",
					IsNullable: false,
					Primary:    true,
				},
				{
					Name:       "website_id",
					Type:       "int(11)",
					IsNullable: false,
					Primary:    true,
				},
				{
					Name:       "value",
					Type:       "varchar(64)",
					IsNullable: false,
					Charset:    "utf8mb4",
					Collation:  "utf8mb4_unicode_ci",
				},
			},
		},
	},
	Indexes: []indexDefinition{
		// PRIMARY KEYS
		/*
			{
				Name:  "",
				Type:  database_ENUM_INDEX_PRIMARY,
				Table: "user",
				Columns: []string{
					"id",
				},
			},
			{
				Name:  "",
				Type:  database_ENUM_INDEX_PRIMARY,
				Table: "folder",
				Columns: []string{
					"id",
				},
			},
			{
				Name:  "",
				Type:  database_ENUM_INDEX_PRIMARY,
				Table: "tag",
				Columns: []string{
					"id",
				},
			},
			{
				Name:  "",
				Type:  database_ENUM_INDEX_PRIMARY,
				Table: "website",
				Columns: []string{
					"id",
				},
			},
			{
				Name:  "",
				Type:  database_ENUM_INDEX_PRIMARY,
				Table: "tags_websites",
				Columns: []string{
					"tag_id",
					"website_id",
				},
			},
		*/

		// UNIQUE KEYS

		{
			Name:  "UNIQUE_user_name",
			Type:  database_ENUM_INDEX_UNIQUE,
			Table: "user",
			Columns: []string{
				"name",
			},
		},
		{
			Name:  "UNIQUE_website_name_folder",
			Type:  database_ENUM_INDEX_UNIQUE,
			Table: "website",
			Columns: []string{
				"name",
				"folder_id",
			},
		},
	},
	ForeignKeys: []foreignKeyDefinition{
		{
			Name:             "FK_tags_websites__tag_id",
			Table:            "tags_websites",
			ColumnName:       "tag_id",
			PointingToTable:  "tag",
			PointingToColumn: "id",
			UpdateRule:       "CASCADE",
			DeleteRule:       "CASCADE",
		},
		{
			Name:             "FK_tags_websites__website_id",
			Table:            "tags_websites",
			ColumnName:       "website_id",
			PointingToTable:  "website",
			PointingToColumn: "id",
			UpdateRule:       "CASCADE",
			DeleteRule:       "CASCADE",
		},
	},
	StoredProcedures: []storedProcedureDefinition{
		{
			Name:    "ListUsers",
			Command: "SELECT id, name, permission_level FROM user;",
		},
	},
	StoredFunctions: []storedFunctionDefinition{
		{
			Name: "CreateUser",
			Parameters: []storedRoutineParameterDefinition{
				{
					Name:      "f_user_name",
					Type:      "VARCHAR(256)",
					Charset:   "utf8mb4",
					Collation: "utf8mb4_unicode_ci",
				},
				{
					Name:      "f_user_password",
					Type:      "VARCHAR(256)",
					Charset:   "utf8mb4",
					Collation: "utf8mb4_unicode_ci",
				},
			},
			Return: storedRoutineParameterDefinition{
				Type: "INT(11)",
			},
			Command: `
DECLARE f_user_id INT(11);
DECLARE f_salt VARCHAR(32) CHARACTER SET utf8mb4 COLLATE utf8mb4_bin;
DECLARE f_password VARCHAR(256) CHARACTER SET utf8mb4 COLLATE utf8mb4_bin;

SELECT TO_BASE64(RANDOM_BYTES(16)) INTO f_salt;
SELECT SHA2(CONCAT(LOWER(f_user_name), f_user_password, f_salt), 512) INTO f_password;

INSERT INTO user(name, password, password_salt) VALUES (f_user_name, f_password, f_salt);

SELECT LAST_INSERT_ID() INTO f_user_id;

RETURN f_user_id;`,
		},
		{
			Name: "CheckUser",
			Parameters: []storedRoutineParameterDefinition{
				{
					Name:      "f_user_name",
					Type:      "VARCHAR(256)",
					Charset:   "utf8mb4",
					Collation: "utf8mb4_unicode_ci",
				},
				{
					Name:      "f_user_password",
					Type:      "VARCHAR(256)",
					Charset:   "utf8mb4",
					Collation: "utf8mb4_bin",
				},
			},
			Return: storedRoutineParameterDefinition{
				Type: "INT(11)",
			},
			Command: `
DECLARE f_user_id INT(11) DEFAULT NULL;
DECLARE f_salt VARCHAR(32) CHARACTER SET utf8mb4 COLLATE utf8mb4_bin;
DECLARE f_password VARCHAR(256) CHARACTER SET utf8mb4 COLLATE utf8mb4_bin;
DECLARE f_user VARCHAR(256) CHARACTER SET utf8mb4 COLLATE utf8mb4_unicode_ci;

SET f_user = LOWER(f_user_name);

SELECT password_salt INTO f_salt FROM user WHERE name = f_user LIMIT 1;
SELECT SHA2(CONCAT(f_user, f_user_password, f_salt), 512) INTO f_password;

SELECT id INTO f_user_id FROM user WHERE name = f_user AND password = f_password LIMIT 1;

RETURN f_user_id;`,
		},
	},
}

var dbPath string
var dbConfig config.Database
var isInit bool
var storedRoutines map[string]uint8
var db *sql.DB

func Init() {
	if isInit {
		return
	}

	dbConfig = config.GetConfig().Database
	dbPath = fmt.Sprintf("%s:%s@(%s:%d)/%s", dbConfig.User, dbConfig.Password, dbConfig.Host, dbConfig.Port, dbConfig.Database)

	profiler.Register(profiler_DATABASE_CONNECT, 200, "µs")
	profiler.Register(profiler_DATABASE_QUERY, 4000, "µs")

	isInit = true
	initDatabaseStructure()
}

func connect() {
	var err error
	log.Printf("Connecting to database %s at %s:%d as %s\n", dbConfig.Database, dbConfig.Host, dbConfig.Port, dbConfig.User)

	tStart := time.Now()
	for i := range dbConfig.RetriesOnError {
		db, err = sql.Open("mysql", dbPath)
		if err == nil {
			break
		}
		log.Printf("Error while connecting the database [%d/%d] : %s\n", i+1, dbConfig.RetriesOnError, err)
		if i+1 == dbConfig.RetriesOnError {
			break
		}
		time.Sleep((time.Duration)(dbConfig.TimeBetweenRetriesMs) * time.Millisecond)
	}

	if err != nil {
		log.Fatalln("Cannot connect to database !")
	}

	db.SetConnMaxLifetime(time.Minute * 1)
	db.SetMaxOpenConns(100)
	db.SetMaxIdleConns(100)

	ping()

	profiler.Add(profiler_DATABASE_CONNECT, time.Duration(time.Since(tStart).Microseconds()))
}

func ExecuteStoredRoutine(routineName string, args ...any) DatabaseResult {
	if !isInit {
		return DatabaseResult{
			Error: createDatabaseError("Database not Init"),
		}
	}
	if storedRoutines == nil {
		fetchStoredRoutines()
	}

	var exist bool
	var routineType uint8
	routineType, exist = storedRoutines[routineName]
	if !exist {
		return DatabaseResult{
			Error: createDatabaseError(fmt.Sprintf("Routine \"%s\" does not exist", routineName)),
		}
	}

	var databaseResult DatabaseResult

	switch routineType {
	case database_ROUTINE_TYPE_PROCEDURE:
		databaseResult = callStoredProcedure(routineName, args...)
	case database_ROUTINE_TYPE_FUNCTION:
		databaseResult = callStoredFunction(routineName, args...)
	default:
		return DatabaseResult{
			Error: createDatabaseError(fmt.Sprintf("Routine \"%s\" isn't a FUNCTION nor a PROCEDURE", routineName)),
		}

	}

	return databaseResult
}

func DecodeDatabaseResult(databaseResult *DatabaseResult) {
	var err error
	var rows *sql.Rows = databaseResult.SqlRows

	if rows != nil {
		defer func() {
			databaseResult.SqlRows.Close()
			databaseResult.SqlRows = nil
		}()
		var columns []string
		if columns, err = rows.Columns(); err != nil {
			databaseResult.Error = createDatabaseError(fmt.Sprintf("Error reading columns name : %s", err))
			rows.Close()
			databaseResult.SqlRows = nil
			return
		}

		if rows.Err() != nil {
			fmt.Printf("Error : %s\n", rows.Err())
		}

		databaseResult.ParsedResult = map[string][]sql.NullString{}
		var numCol int = len(columns)
		for _, c := range columns {
			databaseResult.ParsedResult[c] = make([]sql.NullString, 0)
		}

		for rows.Next() {
			var args []any = make([]any, numCol)
			var strArr []sql.NullString = make([]sql.NullString, numCol)
			for i := range numCol {
				args[i] = &strArr[i]
			}
			if err = rows.Scan(args...); err != nil {
				databaseResult.Error = createDatabaseError(fmt.Sprintf("Error Scanning database response : %s", err))
				break
			}
			for i, c := range columns {
				databaseResult.ParsedResult[c] = append(databaseResult.ParsedResult[c], strArr[i])
			}
		}

		if err = rows.Err(); err != nil {
			databaseResult.Error = createDatabaseError(fmt.Sprintf("Error Scanning database response : %s", err))
		}

	}
}

func ping() {
	var err error
	for i := range dbConfig.RetriesOnError {
		err = db.Ping()
		if err == nil {
			break
		}
		log.Printf("Error while pinging the database [%d/%d] : %s\n", i+1, dbConfig.RetriesOnError, err)
		if i+1 == dbConfig.RetriesOnError {
			break
		}
		time.Sleep((time.Duration)(dbConfig.TimeBetweenRetriesMs) * time.Millisecond)
	}

	if err != nil {
		log.Fatalln("Cannot ping the database !")
	}
}

func executeQuery(query string, args ...any) DatabaseResult {
	if !isInit {
		return DatabaseResult{
			Error: createDatabaseError("Database not Init"),
		}
	}
	if db == nil {
		connect()
	}

	var err error
	var rows *sql.Rows
	tStart := time.Now()
	for i := range dbConfig.RetriesOnError {
		rows, err = db.Query(query, args...)
		if err == nil {
			break
		}

		log.Printf("Error while querying the database [%d/%d] : %s\n", i+1, dbConfig.RetriesOnError, err)
		if i+1 == dbConfig.RetriesOnError {
			break
		}
		time.Sleep((time.Duration)(dbConfig.TimeBetweenRetriesMs) * time.Millisecond)
		ping()
	}
	profiler.Add(profiler_DATABASE_QUERY, time.Duration(time.Since(tStart).Microseconds()))

	if err != nil {
		log.Printf("Cannot query the database ! Query : %s", query)
		log.Printf("With args : %s", args...)
		return DatabaseResult{
			Error: createDatabaseError("Database Query Failed"),
		}
	}

	return DatabaseResult{
		SqlRows: rows,
	}
}

func executeExec(query string, args ...any) DatabaseResult {
	if !isInit {
		return DatabaseResult{
			Error: createDatabaseError("Database not Init"),
		}
	}
	if db == nil {
		connect()
	}

	var err error
	var result sql.Result
	tStart := time.Now()
	for i := range dbConfig.RetriesOnError {
		result, err = db.Exec(query, args...)
		if err == nil {
			break
		}
		log.Printf("Error while executing on the database [%d/%d] : %s\n", i+1, dbConfig.RetriesOnError, err)
		if i+1 == dbConfig.RetriesOnError {
			break
		}
		time.Sleep((time.Duration)(dbConfig.TimeBetweenRetriesMs) * time.Millisecond)
		ping()
	}
	profiler.Add(profiler_DATABASE_QUERY, time.Duration(time.Since(tStart).Microseconds()))

	if err != nil {
		log.Printf("Cannot execute on the database ! Command : %s", query)
		log.Printf("With args : %s\n", args)
		return DatabaseResult{
			Error: createDatabaseError("Database Exec Failed"),
		}
	}

	return DatabaseResult{
		SqlResult: result,
	}
}

func fetchStoredRoutines() {
	storedRoutines = nil
	var dbResult DatabaseResult = executeQuery("SELECT ROUTINE_NAME, ROUTINE_TYPE FROM information_schema.ROUTINES WHERE ROUTINE_SCHEMA=?;", dbConfig.Database)
	var rows *sql.Rows = dbResult.SqlRows
	if rows != nil {
		storedRoutines = make(map[string]uint8, 0)
		defer rows.Close()
		for rows.Next() {
			var resName, resType string
			rows.Scan(&resName, &resType)
			switch resType {
			case "PROCEDURE":
				storedRoutines[resName] = database_ROUTINE_TYPE_PROCEDURE
			case "FUNCTION":
				storedRoutines[resName] = database_ROUTINE_TYPE_FUNCTION
			}
		}
	}
	fmt.Print("ROUTINES : ")
	fmt.Println(storedRoutines)
}

func callStoredProcedure(name string, args ...any) DatabaseResult {
	var query string
	if len(args) == 0 {
		query = fmt.Sprintf("CALL %s()", name)
	} else if len(args) == 1 {
		query = fmt.Sprintf("CALL %s(?)", name)
	} else {
		query = strings.Repeat(", ?", len(args)-1)
		query = fmt.Sprintf("CALL %s(?%s)", name, query)
	}

	var dbResult DatabaseResult = executeQuery(query, args...)
	return dbResult
}

func callStoredFunction(name string, args ...any) DatabaseResult {
	var query string
	if len(args) == 0 {
		query = fmt.Sprintf("SELECT %s() AS result", name)
	} else if len(args) == 1 {
		query = fmt.Sprintf("SELECT %s(?) AS result", name)
	} else {
		query = strings.Repeat(", ?", len(args)-1)
		query = fmt.Sprintf("SELECT %s(?%s) AS result", name, query)
	}

	var dbResult DatabaseResult = executeQuery(query, args...)
	return dbResult
}

func fetchTables() []string {
	var queryResult DatabaseResult = executeQuery("SHOW TABLES;")
	var rows *sql.Rows = queryResult.SqlRows
	//var rows *sql.Rows = executeQuery("SELECT TABLE_NAME FROM information_schema.TABLES WHERE TABLE_SCHEMA=?;", dbConfig.Database)
	var tables []string = make([]string, 0)
	if rows != nil {
		defer rows.Close()
		for rows.Next() {
			var resName string
			rows.Scan(&resName)
			tables = append(tables, resName)
		}
	}
	return tables
}

func fetchColumns(table string) []columsDefinition {
	var queryResult DatabaseResult = executeQuery("SELECT COLUMN_NAME, COLUMN_TYPE, COLUMN_DEFAULT, IS_NULLABLE, CHARACTER_SET_NAME, COLLATION_NAME FROM information_schema.COLUMNS WHERE TABLE_SCHEMA=? AND TABLE_NAME=?;", dbConfig.Database, table)
	var rows *sql.Rows = queryResult.SqlRows
	var columns []columsDefinition = make([]columsDefinition, 0)
	if rows != nil {
		defer rows.Close()
		for rows.Next() {
			var resName, resType, resDefault, resNullable, resCharset, resCollation string
			rows.Scan(&resName, &resType, &resDefault, &resNullable, &resCharset, &resCollation)
			columns = append(columns, columsDefinition{
				Name:       resName,
				Type:       resType,
				Default:    resDefault,
				IsNullable: resNullable == "YES",
				Charset:    resCharset,
				Collation:  resCollation,
				Table:      table,
			})
		}
	}
	return columns
}

func fetchIndexes() []indexDefinition {
	// SELECT * FROM STATISTICS as S INNER JOIN TABLE_CONSTRAINTS as TC ON TC.TABLE_SCHEMA = S.TABLE_SCHEMA AND TC.CONSTRAINT_NAME = S.INDEX_NAME AND TC.TABLE_NAME = S.TABLE_NAME WHERE S.TABLE_SCHEMA = "RPG";
	// SELECT S.COLUMN_NAME as COLUMN_NAME, S.INDEX_NAME AS INDEX_NAME, TC.CONSTRAINT_TYPE AS CONSTRAINT_TYPE FROM STATISTICS as S INNER JOIN TABLE_CONSTRAINTS as TC ON TC.TABLE_SCHEMA = S.TABLE_SCHEMA AND TC.CONSTRAINT_NAME = S.INDEX_NAME AND TC.TABLE_NAME = S.TABLE_NAME WHERE S.TABLE_SCHEMA = "RPG" AND S.TABLE_NAME = "folders" AND TC.CONSTRAINT_TYPE IN("UNIQUE", "PRIMARY KEY");

	// Add PRIMARY : ALTER TABLE <table> ADD CONSTRAINT PRIMARY KEY(<columns>);
	// Add UNIQUE : ALTER TABLE <table> ADD CONSTRAINT <constraint_name> UNIQUE(<columns>);

	var queryResult DatabaseResult = executeQuery("SELECT S.COLUMN_NAME as COLUMN_NAME, S.INDEX_NAME AS INDEX_NAME, TC.CONSTRAINT_TYPE AS CONSTRAINT_TYPE, S.TABLE_NAME AS TABLE_NAME FROM information_schema.STATISTICS as S INNER JOIN information_schema.TABLE_CONSTRAINTS as TC ON TC.TABLE_SCHEMA = S.TABLE_SCHEMA AND TC.CONSTRAINT_NAME = S.INDEX_NAME AND TC.TABLE_NAME = S.TABLE_NAME WHERE S.TABLE_SCHEMA = ? AND TC.CONSTRAINT_TYPE IN(\"UNIQUE\", \"PRIMARY KEY\");", dbConfig.Database)
	var rows *sql.Rows = queryResult.SqlRows
	var indexes []indexDefinition = make([]indexDefinition, 0)
	if rows != nil {
		defer rows.Close()
		for rows.Next() {
			var resColumn, resIndex, resType, resTable string
			rows.Scan(&resColumn, &resIndex, &resType, &resTable)
			var index indexDefinition = indexDefinition{
				Table:   resTable,
				Name:    resIndex,
				Type:    database_ENUM_INDEX_PRIMARY,
				Columns: []string{resColumn},
			}
			if resType == "UNIQUE" {
				index.Type = database_ENUM_INDEX_UNIQUE
			}
			indexes = append(indexes, index)
		}
	}
	return indexes
}

func fetchForeignKeys() []foreignKeyDefinition {

	// ADD ALTER TABLE <table> ADD CONSTRAINT <constraint_name> FOREIGN KEY(<column>) REFERENCES <ref_table>(<ref_column>) ON DELETE <CASCADE / SET NULL> UPDATE <CASCADE / SET NULL>;

	// SELECT RC.CONSTRAINT_NAME as CONTRAINT_NAME, KCU.TABLE_NAME, KCU.COLUMN_NAME, KCU.REFERENCED_TABLE_NAME, KCU.REFERENCED_COLUMN_NAME, RC.UPDATE_RULE, RC.DELETE_RULE FROM REFERENTIAL_CONSTRAINTS AS RC INNER JOIN KEY_COLUMN_USAGE AS KCU ON KCU.CONSTRAINT_NAME = RC.CONSTRAINT_NAME WHERE KCU.TABLE_SCHEMA = "RPG";
	var queryResult DatabaseResult = executeQuery("SELECT RC.CONSTRAINT_NAME as CONTRAINT_NAME, KCU.TABLE_NAME, KCU.COLUMN_NAME, KCU.REFERENCED_TABLE_NAME, KCU.REFERENCED_COLUMN_NAME, RC.UPDATE_RULE, RC.DELETE_RULE FROM information_schema.REFERENTIAL_CONSTRAINTS AS RC INNER JOIN information_schema.KEY_COLUMN_USAGE AS KCU ON KCU.CONSTRAINT_NAME = RC.CONSTRAINT_NAME WHERE KCU.TABLE_SCHEMA = ?;", dbConfig.Database)
	var rows *sql.Rows = queryResult.SqlRows
	var foreignKeys []foreignKeyDefinition = make([]foreignKeyDefinition, 0)
	if rows != nil {
		defer rows.Close()
		for rows.Next() {
			var resName, resTable, resColumn, resRefTable, resRefColumn, resUpdate, resDelete string
			rows.Scan(&resName, &resTable, &resColumn, &resRefTable, &resRefColumn, &resUpdate, &resDelete)
			var fk foreignKeyDefinition = foreignKeyDefinition{
				Name:             resName,
				Table:            resTable,
				ColumnName:       resColumn,
				PointingToTable:  resRefTable,
				PointingToColumn: resRefColumn,
				UpdateRule:       resUpdate,
				DeleteRule:       resDelete,
			}
			foreignKeys = append(foreignKeys, fk)
		}
	}
	return foreignKeys
}

func initDatabaseStructure() {
	fetchStoredRoutines()

	log.Println("Tables :")
	var tables []string = fetchTables()
	for i := range tables {
		var columns []columsDefinition = fetchColumns(tables[i])
		log.Printf(" - Table %s : \n", tables[i])
		for j := range columns {
			log.Printf("   - Column %s %s IsNullable:%t Default:%s Charset:%s/%s\n",
				columns[j].Name, columns[j].Type, columns[j].IsNullable, columns[j].Default, columns[j].Charset, columns[j].Collation)
		}
	}

	log.Println("Indexes :")
	var indexes []indexDefinition = fetchIndexes()
	for i := range indexes {
		log.Printf(" - Index %s\n", indexes[i])
	}

	log.Println("Foreign keys :")
	var foreignKeys []foreignKeyDefinition = fetchForeignKeys()
	for i := range foreignKeys {
		log.Printf(" - Foreign key %s\n", foreignKeys[i])
	}
}

func UpdateDatabaseStructure() {
	var tablesNeeded []tableDefinition = databaseStructure.Tables

	for t := range tablesNeeded {
		var table tableDefinition = tablesNeeded[t]
		var columnsNeeded []columsDefinition = table.Columns

		var query string = fmt.Sprintf("CREATE TABLE %s (\n", table.Name)
		var primaryKeys []string = make([]string, 0)
		for c := range columnsNeeded {
			if c != 0 {
				query += ",\n"
			}
			var column columsDefinition = columnsNeeded[c]
			query += fmt.Sprintf("%s %s", column.Name, column.Type)

			if str, ok := column.Charset.(string); ok {
				query += fmt.Sprintf(" CHARACTER SET %s", str)
			}

			if str, ok := column.Collation.(string); ok {
				query += fmt.Sprintf(" COLLATE %s", str)
			}

			if !column.IsNullable {
				query += " NOT NULL"
			}

			if str, ok := column.Default.(string); ok {
				query += fmt.Sprintf(" DEFAULT %s", str)
			}

			if column.AutoIncrement {
				query += " AUTO_INCREMENT"
			}

			if column.AutoIncrement || column.Primary {
				primaryKeys = append(primaryKeys, column.Name)
			}
		}

		if len(primaryKeys) > 0 {
			query += fmt.Sprintf(",\nPRIMARY KEY (%s)", strings.Join(primaryKeys, ", "))
		}

		query += "\n);"
		fmt.Println(query)
		executeExec(query)
	}

	var indexesNeeded []indexDefinition = databaseStructure.Indexes

	for i := range indexesNeeded {
		var index indexDefinition = indexesNeeded[i]
		// ALTER TABLE <table> ADD CONSTRAINT PRIMARY KEY(<columns>);
		// ALTER TABLE <table> ADD CONSTRAINT <constraint_name> UNIQUE(<columns>);
		var command string = fmt.Sprintf("ALTER TABLE %s ADD CONSTRAINT", index.Table)
		switch index.Type {
		case database_ENUM_INDEX_PRIMARY:
			command += " PRIMARY KEY("
		case database_ENUM_INDEX_UNIQUE:
			command += fmt.Sprintf(" %s UNIQUE(", index.Name)
		}

		for c := range index.Columns {
			if c != 0 {
				command += ", "
			}
			var column string = index.Columns[c]
			command += column
		}

		command += ");"
		fmt.Println(command)
		executeExec(command)
	}

	var foreignKeys []foreignKeyDefinition = databaseStructure.ForeignKeys

	for fk := range foreignKeys {
		var foreignKey foreignKeyDefinition = foreignKeys[fk]
		// ALTER TABLE <table> ADD CONSTRAINT <constraint_name> FOREIGN KEY(<column>) REFERENCES <ref_table>(<ref_column>) ON DELETE <CASCADE / SET NULL> UPDATE <CASCADE / SET NULL>;
		var command string = fmt.Sprintf("ALTER TABLE %s ADD INDEX %s(%s);", foreignKey.PointingToTable, foreignKey.Name, foreignKey.PointingToColumn)

		fmt.Println(command)
		executeExec(command)

		command = fmt.Sprintf("ALTER TABLE %s ADD CONSTRAINT %s FOREIGN KEY (%s) REFERENCES %s (%s)", foreignKey.Table, foreignKey.Name, foreignKey.ColumnName, foreignKey.PointingToTable, foreignKey.PointingToColumn)

		if foreignKey.DeleteRule != "" {
			command += fmt.Sprintf(" ON DELETE %s", foreignKey.DeleteRule)
		}

		if foreignKey.UpdateRule != "" {
			command += fmt.Sprintf(" ON UPDATE %s", foreignKey.UpdateRule)
		}

		command += ";"

		fmt.Println(command)
		executeExec(command)
	}

	var storedFunctions []storedFunctionDefinition = databaseStructure.StoredFunctions

	for _, f := range storedFunctions {
		// CREATE DEFINER='<user>'@'<host>' FUNCTION <func_name>(<param1>, <param2>, ...) RETURNS <return_type> BEGIN
		//  [...]
		// END

		var command string = fmt.Sprintf("CREATE DEFINER=`%s`@`%s` FUNCTION `%s`(", dbConfig.User, dbConfig.Host, f.Name)

		for i, fp := range f.Parameters {
			if i != 0 {
				command += ", "
			}
			command += fmt.Sprintf("`%s` %s", fp.Name, fp.Type)

			if fpc, ok := fp.Charset.(string); ok {
				command += fmt.Sprintf(" CHARACTER SET %s", fpc)
			}
			if fpc, ok := fp.Collation.(string); ok {
				command += fmt.Sprintf(" COLLATE %s", fpc)
			}
		}
		command += fmt.Sprintf(") RETURNS %s", f.Return.Type)

		if frc, ok := f.Return.Charset.(string); ok {
			command += fmt.Sprintf(" CHARACTER SET %s", frc)
		}
		if frc, ok := f.Return.Collation.(string); ok {
			command += fmt.Sprintf(" COLLATE %s", frc)
		}

		command += fmt.Sprintf(" BEGIN %s END", f.Command)

		fmt.Println(command)
		executeExec(command)
	}

	var storedProcedures []storedProcedureDefinition = databaseStructure.StoredProcedures

	for _, p := range storedProcedures {
		// CREATE DEFINER='<user>'@'<host>' PROCEDURE <proc_name>(IN <param1>, IN <param2>, ...) BEGIN
		//  [...]
		// END

		var command string = fmt.Sprintf("CREATE DEFINER=`%s`@`%s` PROCEDURE `%s`(", dbConfig.User, dbConfig.Host, p.Name)

		for i, pp := range p.Parameters {
			if i != 0 {
				command += ", "
			}
			command += fmt.Sprintf("IN '%s' %s", pp.Name, pp.Type)

			if ppc, ok := pp.Charset.(string); ok {
				command += fmt.Sprintf(" CHARACTER SET %s", ppc)
			}
			if ppc, ok := pp.Collation.(string); ok {
				command += fmt.Sprintf(" COLLATE %s", ppc)
			}
		}
		command += fmt.Sprintf(") BEGIN %s END", p.Command)

		fmt.Println(command)
		executeExec(command)
	}
}

func DestroyDatabase() {
	var tables []string = fetchTables()
	var foreignKeys []foreignKeyDefinition = fetchForeignKeys()

	for fk := range foreignKeys {
		var foreignKey foreignKeyDefinition = foreignKeys[fk]
		var command string = fmt.Sprintf("ALTER TABLE %s DROP FOREIGN KEY %s;", foreignKey.Table, foreignKey.Name)

		fmt.Println(command)
		executeExec(command)
	}

	for t := range tables {
		var table string = tables[t]
		var command string = fmt.Sprintf("DROP TABLE %s;", table)

		fmt.Println(command)
		executeExec(command)
	}

	fetchStoredRoutines()

	for r, t := range storedRoutines {
		var command = "DROP"
		switch t {
		case database_ROUTINE_TYPE_FUNCTION:
			command += " FUNCTION "
		case database_ROUTINE_TYPE_PROCEDURE:
			command += " PROCEDURE "
		}
		command += r

		fmt.Println(command)
		executeExec(command)
	}
}
